/******************************************************************************

cc snow.c -o snow -I/usr/X11R6/include -L/usr/X11R6/lib -lGL -lGLU -lglut -lm

Usage: ./snow basename imagesize nframes
Generates multiple frames in basenameNNNNN.ppm, etc.
Imagesize is 256, 128, etc.

Generates a field of spheres and moves camera up with constant velocity.
(Simulates particles falling with constant velocity.)

TODO:
- what kind of texture do we want?
- antialiasing
- Note: get frame corruption on glReadPixels() if window in front of us.

Richard Mann, School of Computer Science, University of Waterloo.

******************************************************************************/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <GL/glut.h>
#include <math.h>       /* for cos(), sin(), and sqrt() */

#ifndef M_PI
#define M_PI 3.14159265
#endif

#if 0  /* drand48 might be better on Unix machines */
#define RANDOM_RANGE(lo, hi) ((lo) + (hi - lo) * drand48())
#else
static float float_rand(void) { return rand() / (float) RAND_MAX; }
#define RANDOM_RANGE(lo, hi) ((lo) + (hi - lo) * float_rand())
#endif

#define DVY 0.025 /* motion in Y direction */
#define MAXFRAMES 128 /* number of frames, power of two */
#define NP 4000
#define LX 4.0 /* size of volume; extrude along Y direction */
#define LZ 8.0 /* more depth of field */
#define LY (LX+MAXFRAMES*DVY)
float point[NP][3];
#define VZ 10.0 /* camera position */
#define FOV 30.0 /* field of view in degrees */
#define PN 1.0 /* near and far clipping planes */
#define PF 30.0

float vx=0.0;  float vy=0.0;  float vz=VZ; /* viewer position */
float dvx=0.0; float dvy=DVY; float dvz=0.0; /* viewer motion */
int frame_count=0; /* frame number */
GLubyte *frameBuffer;
char basename[100]; /* base filename */
int imagesize, width, height;
int nframes;

#define BIGVAL 9999.0
#define DIAMETER 0.2

float sqr(float x) { return x*x; }

float dist(float a[], float b[]) {
  float d = sqr(a[0]-b[0]) + sqr(a[1]-b[1]) + sqr(a[2]-b[2]);
  return sqrt(d);
}

void init(void) {
  int i, j;
  /* Graphics people can worry about these... */
  GLfloat mat_specular[] = { 1.0, 1.0, 1.0, 1.0 };
  GLfloat mat_shininess[] = { 50.0 };
  GLfloat light_position[] = { 0.0, 0.0, 10.0, 0.0 }; /* at viewer */

  srand((unsigned int)time(NULL));
  /* srand((unsigned int) glutGet(GLUT_ELAPSED_TIME) ); */

  glClearColor(0.0, 0.0, 0.0, 0.0);
  glShadeModel(GL_SMOOTH);
  glEnable(GL_DEPTH_TEST);

  /* glLightfv(GL_LIGHT0, GL_AMBIENT, ambient); */
  /* glLightfv(GL_LIGHT0, GL_DIFFUSE, mat_specular);*/
  glLightfv(GL_LIGHT0, GL_POSITION, light_position);
  glEnable(GL_LIGHTING);
  glEnable(GL_LIGHT0);
  
  /* make random points here */
  for (i=0; i<NP; i++) {
    point[i][0]=RANDOM_RANGE(-LX,LX);
    point[i][1]=RANDOM_RANGE(-LY,LY); /* Extrude along Y direction */
    point[i][2]=RANDOM_RANGE(-LZ,LZ);
    for (j=0; j<i; j++) {
      if (dist(point[i],point[j]) < 2*DIAMETER) {
	printf("point %d: (%f,%f,%f), %d: (%f,%f,%f)\n",
	       i,point[i][0],point[i][1],point[i][2],
	       j,point[j][0],point[j][1],point[j][2]);
	point[i][0]=BIGVAL;
	break;
      }}
  }
}

void display(void) {
  int i;
  float x, y, z;
  int rsize;
  GLubyte *rptr1, *rptr2;
  GLubyte temp;
  char tempfn[100];
  FILE *fp;

  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

  glLoadIdentity ();             /* clear the matrix */

  /* draw spheres centered at points, restoring coords each time */
  for (i=0; i<NP; i++) {
    glPushMatrix();
    x=point[i][0]; y=point[i][1]; z=point[i][2];
    if (x<BIGVAL) {
      glTranslatef(x, y, z);
      glutSolidSphere(DIAMETER, 20, 16); /* 0.2 diameter */
      glPopMatrix();
    }
  }
  glutSwapBuffers();

  printf("Frame:%d, view:(%f,%f,%f)\n", frame_count, vx, vy, vz);
  /* Collect contents of frame buffer */
  glReadBuffer(GL_FRONT);
  width=imagesize; height=imagesize;
  glReadPixels(0, 0, width, height, GL_RGB, GL_UNSIGNED_BYTE, frameBuffer);

  /* Flip in Y direction */
  rsize = width*3;
  rptr1 = (GLubyte *)&frameBuffer[0];
  rptr2 = (GLubyte *)&frameBuffer[rsize*(height-1)];
  while (rptr1 < rptr2) {
    for (i=0; i<rsize; i++) {
      temp = *(rptr1+i);
      *(rptr1+i) = *(rptr2+i);
      *(rptr2+i) = temp; }
    rptr1+=rsize; rptr2-=rsize; }

  /* Write out PPM file */
  sprintf(tempfn, "%s%05d.ppm", basename, frame_count);
  fp=fopen(tempfn, "w");
  fprintf(fp, "P6\n%d\n%d\n255\n", width, height);
  fwrite(frameBuffer, width*height*3, 1, fp);
  fclose(fp);
  frame_count++;
  if (frame_count == nframes) exit(0);

}

void reshape (int w, int h) {
  float t=20.0*M_PI/180.0;
  width=w; height=h;
  glViewport(0, 0, (GLsizei) width, (GLsizei) height); 
  /* Allocate image buffer each time screen size changes! */
  /* This *should* be called before any drawing is done. */
  frameBuffer = (GLubyte *)malloc(width*height*3);
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  /* 30 degree field of view, aspect=1, near=1, far= 11 */
  gluPerspective(FOV, 1.0, PN, PF);
  /* Looking down negative Z axis from [vx,vy,vz], up is (0,1,0) */
  /* gluLookAt(vx, vy, vz, vx, vy, -100.0, 0.0, 1.0, 0.0); */
  gluLookAt(vx, vy, vz, vx, vy, -100.0, sin(t), cos(t), 0.0);
  glMatrixMode(GL_MODELVIEW);
}

void moveCamera () {
  float t=20.0*M_PI/180.0;
  vx+=dvx; vy+=dvy; vz+=dvz;

  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  /* 30 degree field of view, near=1, far= 11 */
  gluPerspective(FOV, 1.0, PN, PF);
  /* Looking down negative Z axis from [vx,vy,vz], up is (0,1,0) */
  gluLookAt(vx, vy, vz, vx, vy, -100.0, sin(t), cos(t), 0.0);
  glMatrixMode(GL_MODELVIEW);

  glutPostRedisplay();
}

/* ARGSUSED1 */
void keyboard(unsigned char key, int x, int y) {
  switch (key) {
  case 27:
    exit(0);
    break;
  }
}

int main(int argc, char** argv) {
  if (argc<3) {
    fprintf(stderr, "Usage: %s basename size nframes\n", argv[0]); exit(-1); }
  sprintf(basename, "%s", argv[1]);
  imagesize=atoi(argv[2]);
  nframes=atoi(argv[3]);
 
  glutInit(&argc, argv);
  glutInitDisplayMode(GLUT_RGB | GLUT_DEPTH | GLUT_DOUBLE);
  width=imagesize; height=imagesize;
  glutInitWindowSize (width, height); 
  glutInitWindowPosition (100, 100);
  glutCreateWindow (argv[0]);
  init ();
  glutDisplayFunc(display); 
  glutReshapeFunc(reshape);
  glutKeyboardFunc(keyboard);
  glutIdleFunc(moveCamera);
  glutMainLoop();
  return 0;
}
